// Copyright Daniel Wallin, David Abrahams 2005. Use, modification and
// distribution is subject to the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//
// Modified by Roland Pihlakas (2007) 
//

#ifndef ARG_LIST_050329_HPP
#define ARG_LIST_050329_HPP

#include <boost/parameter/aux_/void.hpp>
#include <boost/parameter/aux_/result_of0.hpp>
#include <boost/parameter/aux_/default.hpp>
#include <boost/parameter/aux_/parameter_requirements.hpp>
#include <boost/parameter/config.hpp>

#include <boost/mpl/apply.hpp>

#include <boost/type_traits/add_reference.hpp>
#include <boost/type_traits/is_same.hpp>

#include <boost/preprocessor/repetition/enum_params.hpp>
#include <boost/preprocessor/facilities/intercept.hpp>

namespace boost { namespace parameter { 

// Forward declaration for aux::arg_list, below.
template<class T> struct keyword;

namespace aux {



//roland start - allows merging with empty argument lists

template <class TaggedArg, class Next>
struct arg_list;

template <class KW, class T>
struct tagged_argument;

//roland end - allows merging with empty argument lists



//
// Structures used to build the tuple of actual arguments.  The
// tuple is a nested cons-style list of arg_list specializations
// terminated by an empty_arg_list.
//
// Each specialization of arg_list is derived from its successor in
// the list type.  This feature is used along with using
// declarations to build member function overload sets that can
// match against keywords.
//
  
// Terminates arg_list<> and represents an empty list.  Since this
// is just the terminating case you might want to look at arg_list
// first, to get a feel for what's really happening here.
struct empty_arg_list
{
    inline empty_arg_list() {}

    // Constructor taking BOOST_PARAMETER_MAX_ARITY empty_arg_list
    // arguments; this makes initialization
    empty_arg_list(
        BOOST_PP_ENUM_PARAMS(
            BOOST_PARAMETER_MAX_ARITY, void_ BOOST_PP_INTERCEPT
        ))
    {}

    // A metafunction class that, given a keyword and a default
    // type, returns the appropriate result type for a keyword
    // lookup given that default
    struct binding
    {
        template<class KW, class Default>
        struct apply
        {
            typedef Default type;
        };
    };
    
    
    
    //roland start - useful when we want to read some typedefs in argument's class

    struct binding_noref
    {
        template<class KW, class Default>
        struct apply
        {
            typedef Default type;
        };
    };

    //roland end - useful when we want to read some typedefs in argument's class
    
    
    
    
    
    
    
    

#if BOOST_WORKAROUND(BOOST_MSVC, <= 1300) \
    || (BOOST_WORKAROUND(__GNUC__, < 3)) \
    || BOOST_WORKAROUND(__BORLANDC__, BOOST_TESTED_AT(0x564))
    
    // The overload set technique doesn't work with these older
    // compilers, so they need some explicit handholding.
      
    // A metafunction class that, given a keyword, returns the type
    // of the base sublist whose get() function can produce the
    // value for that key
    struct key_owner
    {
        template<class KW>
        struct apply
        {
            typedef empty_arg_list type;
        };
    };

    template <class K, class T>
    T& get(default_<K,T> x) const
    {
        return x.value;
    }

    template <class K, class F>
    typename result_of0<F>::type
    get(lazy_default<K,F> x) const
    {
        return x.compute_default();
    }
#endif

    // If this function is called, it means there is no argument
    // in the list that matches the supplied keyword. Just return
    // the default value.
    template <class K, class Default>
    Default& operator[](default_<K, Default> x) const
    {
        return x.value;
    }

    // If this function is called, it means there is no argument
    // in the list that matches the supplied keyword. Just evaluate
    // and return the default value.
    template <class K, class F>
    inline typename result_of0<F>::type
    operator[](
        BOOST_PARAMETER_lazy_default_fallback<K,F> x) const
    {
        return x.compute_default();
    }

    // No argument corresponding to ParameterRequirements::key_type
    // was found if we match this overload, so unless that parameter
    // has a default, we indicate that the actual arguments don't
    // match the function's requirements.
    template <class ParameterRequirements>
    static typename ParameterRequirements::has_default
    satisfies(ParameterRequirements*);
    
    

    //roland start - if new param has same key as existing somewhere in current list, then we need to find it in order to remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous

	template <class key>
	struct index_of_key
	{
		static const int rval = -1;	  	
	};	
	
	template <class key>
	struct index_of_keyword
	{
		static const int rval = -1;	  	
	};	
	
	
    /*
	template <int index>
	empty_arg_list
	remove_key_from_index_O()
	{
		BOOST_STATIC_ASSERT(index);				//should not be called - if called then index was somehow too big
		return *this;	
	};
	*/
	
	
	
	template <int index>
	struct remove_key_from_index_T
	{
		typedef empty_arg_list type;
	};
    
    
    //roland end - if new param has same key as existing somewhere in current list, then we need to find it in order to remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
    
    
    
    
    
    //roland start - allows merging with empty argument lists
        
        
	template <class KW, class T2>   
	struct add_param
	{
		typedef arg_list<tagged_argument<KW, T2>, empty_arg_list>				type;
	};    
        
/*        
	template <class KW>
	struct typeof
	{
		BOOST_STATIC_ASSERT(false);
	};
*/        
        
        
        
        
	template <class KW, class T2>							//if previous argument list was empty. MSVC compiles without it, but then called function gets tagged_argument type as parameter, not arg_list type
    inline typename add_param<KW, T2>::type 
    operator,(tagged_argument<KW,T2> x) const				//roland - existing argument list is always const
    {
        return arg_list<tagged_argument<KW,T2>, empty_arg_list>(x, empty_arg_list());
    }      
    
    
    
	//this method is useful for case when some macro sometimes adds a parameter and sometimes not (using then empty_arg_list instead), but we still want to use comma before that macro
	inline const empty_arg_list&																 //inline by roland
    operator,(empty_arg_list x) const				//roland - existing argument list is always const
    {   
        return *this;		  
    }
    
    
    
    
        
    //roland end - allows merging with empty argument lists
    
    
    
    
    //roland start - isset, unset and set methods
    
    template <class key>
	bool isset2(key &) const
	{
		return false;
	}
	
	template <class key>
	struct isset
	{
		static const bool rval = false;
	};
	
    
    
    template <class key>
	empty_arg_list unset(/*key &*/) const
	{
		return *this;
	}

	
	/*
	template <class newlist>
	newlist operator=(newlist const &in) const
	{
		return in;
	}
	*/
    
    //roland end - isset, unset and set methods
    
    
};

// Forward declaration for arg_list::operator,
template <class KW, class T>
struct tagged_argument;




//roland start - if new param has same key as existing somewhere in current list, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous

template <class TaggedArg, class Next>
struct arg_list;



#if 0

		template <class arg_list_T, int index>
		struct remove_key_from_index_T_worker
		{
			typedef arg_list<
						typename arg_list_T::TaggedArg_T
					  , typename remove_key_from_index_T_worker<typename arg_list_T::Next_T, index - 1>::type
					>
					type;
		};

		template <class arg_list_T>
		struct remove_key_from_index_T_worker<arg_list_T, 0>
		{
			typedef typename arg_list_T::Next_T type;
		};	

		template <class arg_list_T>
		struct remove_key_from_index_T_worker<arg_list_T, -1>
		{
			typedef arg_list_T type;
		};

#else						 //roland - lets optimize so that new key and value are substituted into old one's memory slot. This way we prevent shifting the whole list around in memory

	#if 0

		template <int index, class new_TaggedArg, class arg_list_T>
		struct replace_key_from_index_T_worker
		{
			typedef arg_list<
						typename arg_list_T::TaggedArg_T
					  , typename replace_key_from_index_T_worker<index - 1, new_TaggedArg, typename arg_list_T::Next_T>::type
					>
					type;
		};

		template <class new_TaggedArg, class arg_list_T>
		struct replace_key_from_index_T_worker<0, new_TaggedArg, arg_list_T>		  //replace takes place here
		{
			//typedef typename arg_list_T::Next_T type;
			typedef arg_list<
						new_TaggedArg
					  , typename arg_list_T::Next_T
					>
					type;
		};	

		template <class new_TaggedArg, class arg_list_T>							   //key was not found, so lets add it
		struct replace_key_from_index_T_worker<-1, new_TaggedArg, arg_list_T>
		{
			//typedef arg_list_T type;
			typedef arg_list<new_TaggedArg, arg_list_T> type;
		};
		
	#endif

#endif

//roland end - if new param has same key as existing somewhere in current list, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous




#pragma warning (push)								    //roland
#pragma warning (disable : 4512)						//roland - warning C4512: 'boost::parameter::aux::arg_list<TaggedArg,Next>' : assignment operator could not be generated	    

// A tuple of tagged arguments, terminated with empty_arg_list.
// Every TaggedArg is an instance of tagged_argument<>.
template <class TaggedArg, class Next = empty_arg_list>
struct arg_list : Next
{
    typedef arg_list<TaggedArg,Next> self;
    typedef typename TaggedArg::key_type key_type;
    typedef typename TaggedArg::value_type value_type;
    typedef typename TaggedArg::reference reference;

    TaggedArg arg;      // Stores the argument

    // Store the arguments in successive nodes of this list
    template< // class A0, class A1, ...
        BOOST_PP_ENUM_PARAMS(BOOST_PARAMETER_MAX_ARITY, class A)
    >
    arg_list( // A0 const& a0, A1 const& a1, ...
        BOOST_PP_ENUM_BINARY_PARAMS(BOOST_PARAMETER_MAX_ARITY, A, const & a)
    )
      : Next( // a1, a2, ...
            BOOST_PP_ENUM_SHIFTED_PARAMS(BOOST_PARAMETER_MAX_ARITY, a)
          , void_()
        )
      , arg(a0)
    {}

    // Create a new list by prepending arg to a copy of tail.  Used
    // when incrementally building this structure with the comma
    // operator.
    inline arg_list(TaggedArg arg, Next const& tail)
      : Next(tail)
      , arg(arg)
    {}


    // A metafunction class that, given a keyword and a default
    // type, returns the appropriate result type for a keyword
    // lookup given that default
    struct binding
    {
        template <class KW, class Default>
        struct apply
        {
          typedef typename mpl::eval_if<
                boost::is_same<KW, key_type>
              , mpl::identity<reference>
              , mpl::apply_wrap2<typename Next::binding, KW, Default>
          >::type type;
        };
    };
    
    
    
    //roland start - useful when we want to read some typedefs in argument's class
    
    struct binding_noref
    {
        template <class KW, class Default>
        struct apply
        {
          typedef typename mpl::eval_if<
                boost::is_same<KW, key_type>
              , mpl::identity<value_type>
              , mpl::apply_wrap2<typename Next::binding_noref, KW, Default>
          >::type type;
        };
    };
    
    //roland end - useful when we want to read some typedefs in argument's class
    
    

    //
    // Begin implementation of indexing operators for looking up
    // specific arguments by name
    //

#if BOOST_WORKAROUND(BOOST_MSVC, <= 1300) \
    || BOOST_WORKAROUND(__GNUC__, < 3) \
    || BOOST_WORKAROUND(__BORLANDC__, BOOST_TESTED_AT(0x564))
    // These older compilers don't support the overload set creation
    // idiom well, so we need to do all the return type calculation
    // for the compiler and dispatch through an outer function template

    // A metafunction class that, given a keyword, returns the base
    // sublist whose get() function can produce the value for that
    // key.
    struct key_owner
    {
        template<class KW>
        struct apply
        {
          typedef typename mpl::eval_if<
                boost::is_same<KW, key_type>
              , mpl::identity<arg_list<TaggedArg,Next> >
              , mpl::apply_wrap1<typename Next::key_owner,KW>
          >::type type;
        };
    };

    // Outer indexing operators that dispatch to the right node's
    // get() function.
    template <class KW>
    typename mpl::apply_wrap2<binding, KW, void_>::type
    operator[](keyword<KW> const& x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *this;
        return sublist.get(x);
    }

    template <class KW, class Default>
    typename mpl::apply_wrap2<binding, KW, Default&>::type
    operator[](default_<KW, Default> x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *this;
        return sublist.get(x);
    }

    template <class KW, class F>
    typename mpl::apply_wrap2<
        binding,KW
      , typename result_of0<F>::type
    >::type
    operator[](lazy_default<KW,F> x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *this;
        return sublist.get(x);
    }

    // These just return the stored value; when empty_arg_list is
    // reached, indicating no matching argument was passed, the
    // default is returned, or if no default_ or lazy_default was
    // passed, compilation fails.
    reference get(keyword<key_type> const&) const
    {
        return arg.value;
    }

    template <class Default>
    reference get(default_<key_type,Default>) const
    {
        return arg.value;
    }

    template <class Default>
    reference get(lazy_default<key_type, Default>) const
    {
        return arg.value;
    }
    
#else

    inline reference operator[](keyword<key_type> const&) const			 //inline by roland
    {
        return arg.value;
    }

    template <class Default>
    inline reference operator[](default_<key_type, Default>) const		 //inline by roland
    {
        return arg.value;
    }

    template <class Default>
    reference operator[](lazy_default<key_type, Default>) const
    {
        return arg.value;
    }

    // Builds an overload set including operator[]s defined in base
    // classes.
    using Next::operator[];

    //
    // End of indexing support
    //





	//roland start

/*
	template <class Keyword2>
	struct typeof
	{
		BOOST_STATIC_ASSERT(false);
	};
	
	template <>
	struct typeof<key_type>
	{
		typedef value_type					type;
	};
*/
	
	//roland end







    //
    // For parameter_requirements matching this node's key_type,
    // return a bool constant wrapper indicating whether the
    // requirements are satisfied by TaggedArg.  Used only for
    // compile-time computation and never really called, so a
    // declaration is enough.
    //
    template <class HasDefault, class Predicate>
    static typename mpl::apply1<Predicate, value_type>::type
    satisfies(
        parameter_requirements<key_type,Predicate,HasDefault>*
    );

    // Builds an overload set including satisfies functions defined
    // in base classes.
    using Next::satisfies;
#endif



#if 0

    // Comma operator to compose argument list without using parameters<>.
    // Useful for argument lists with undetermined length.
    template <class KW, class T2>
    inline arg_list<tagged_argument<KW, T2>, self>						//inline by roland
    operator,(tagged_argument<KW,T2> x)
    {
        return arg_list<tagged_argument<KW,T2>, self>(x, *this);
    }
    
#endif    
    
    
    
    
    
    
    
#if 0    
    
    //roland start - allows merging argument lists
    
	template <class KW, class T2>
    arg_list<tagged_argument<KW, T2>, self> 
    operator,(tagged_argument<KW,T2> x) const				//roland - existing argument list is always const
    {
        return arg_list<tagged_argument<KW,T2>, self>(x, *this);
    }
    
    //roland end - allows merging argument lists
    
    
    
    //roland start - if new param has same key as current list head, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
    
	template <class T2>
    arg_list<tagged_argument<key_type, T2>, Next> 
    operator,(tagged_argument<key_type,T2> x) const				//roland - existing argument list is always const
    {
        return arg_list<            
            tagged_argument<key_type, T2> 
          , Next
        >(
            x
          , static_cast<Next>(*this)
        );
    }
   
    //roland end - if new param has same key as current list head, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
    
#else



    //roland start - if new param has same key as existing somewhere in current list, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous

	template <class key>
	struct index_of_key												//if new param has same key as existing somewhere in current list, then we need to find it in order to remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
	{
		static const int temp = Next::index_of_key<key>::rval;
		static const int rval = (temp == -1) ? -1 : temp + 1;	  	
	};	
	
	template <>
	struct index_of_key<key_type>
	{
		static const int rval = 0;
	};
	
	
	
	
	
	template <class key>
	struct index_of_keyword												//if new param has same key as existing somewhere in current list, then we need to find it in order to remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
	{
		static const int temp = Next::index_of_keyword<key>::rval;
		static const int rval = (temp == -1) ? -1 : temp + 1;	  	
	};	
	
	template <>
	struct index_of_keyword<keyword<key_type> >
	{
		static const int rval = 0;
	};
	
	
	
	
	
	//typedef Next		Next_T;
	//typedef TaggedArg	TaggedArg_T;
	
	
	
	
	
		#if 0	
			
			template <int index>
			struct remove_key_from_index_T
			{
				typedef typename remove_key_from_index_T_worker<self, index>::type type;			
			};


			
			template <int index>
			inline typename remove_key_from_index_T<index>::type				 //inline by roland
			remove_key_from_index_O() const							 	
			{
				return 
				remove_key_from_index_T<index>::type
				(
					arg
				  , Next::remove_key_from_index_O<index - 1>()	
				);
			}
			
			template <>
			inline Next															 //inline by roland
			remove_key_from_index_O<0>() const
			{
				return static_cast<Next>(*this);
			}	
			
			template <>
			inline self															  //inline by roland
			remove_key_from_index_O<-1>() const
			{
				return *this;
			}
			
		#else	 //roland - lets optimize so that new key and value are substituted into old one's memory slot. This way we prevent shifting the whole list around in memory






#if 0

	template <int index, class new_TaggedArg>
	struct replace_key_from_index_T
	{
		typedef typename replace_key_from_index_T_worker<index, new_TaggedArg, self>::type type;			
	};


	
	template <int index, class new_TaggedArg, class oldlist_T>								//partial function template specialization is not allowed therefore we need to do this dance
	struct do_replace_key_from_index_O
	{
		static typename replace_key_from_index_T_worker<index, new_TaggedArg, oldlist_T>::type
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)							 	
		{
			return /*arg_list<
					typename oldlist_T::TaggedArg_T
				  , replace_key_from_index_T_worker<index - 1, new_TaggedArg, typename oldlist_T::Next_T>::type
				>	 */
				replace_key_from_index_T_worker<index, new_TaggedArg, oldlist_T>::type
				(
					oldlist.arg
				  , do_replace_key_from_index_O
					<
						index - 1
					  , new_TaggedArg
					  , typename oldlist_T::Next_T
					>
					::apply													//++
					(
						new_arg
					  , static_cast<typename oldlist_T::Next_T>(oldlist)
					)	
				);
		}
	};
	
	
	template <class new_TaggedArg, class oldlist_T>
	struct do_replace_key_from_index_O<0, new_TaggedArg, oldlist_T>
	{
		BOOST_STATIC_ASSERT(false);
	
		static arg_list<new_TaggedArg, typename oldlist_T::Next_T>
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)				//replace takes place here
		{
			//return static_cast<Next>(*this);
			return arg_list<            
				new_TaggedArg 
			  , typename oldlist_T::Next_T
			>(
				new_arg
			  , static_cast<typename oldlist_T::Next_T>(oldlist)
			);
		}	
	};
	
	
	template <class new_TaggedArg, class oldlist_T>
	struct do_replace_key_from_index_O<-1, new_TaggedArg, oldlist_T>
	{
		static arg_list<new_TaggedArg, self>
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)				//key was not found, so lets add it
		{
			//return *this;
			return arg_list<new_TaggedArg, self>(new_arg, oldlist);
		}
	};
	
#else

	/*
	template <int index, class new_TaggedArg>
	struct replace_key_from_index_T_worker;
	
	
	template <int index, class new_TaggedArg>
	struct replace_key_from_index_T
	{
		typedef typename replace_key_from_index_T_worker<index, new_TaggedArg>::type type;			
	};
	*/
	
	
	
	template <int index>
	struct remove_key_from_index_T
	{
		/*template <int index2>
		struct go
		{
			typedef arg_list<
						TaggedArg
					  , Next_T::remove_key_from_index_T<index2>::type> type;
		};
		
		typedef typename go<index - 1>::type		type;*/
		
		typedef Next		Next_T;
		
		typedef arg_list<
					  TaggedArg
					, typename Next_T::remove_key_from_index_T<index - 1>::type
				  > 
				  type;
	};
	
	template <>
	struct remove_key_from_index_T<1/*0*/>
	{
		typedef Next type;
	};	
	
	template <>
	struct remove_key_from_index_T<0/*-1*/>
	{
		typedef self type;
	};	


	
	

	template <int index, class new_TaggedArg/*, class arg_list_T*/>
	struct replace_key_from_index_T
	{
		typedef Next		Next_T;
	
		typedef arg_list<
					TaggedArg
				  , typename Next_T::replace_key_from_index_T<index - 1, new_TaggedArg>::type
				>
				type;
	};

	template <class new_TaggedArg/*, class arg_list_T*/>
	struct replace_key_from_index_T<1/*0*/, new_TaggedArg/*, arg_list_T*/>		  //replace takes place here
	{
		//typedef typename arg_list_T::Next_T type;
		typedef arg_list<
					new_TaggedArg
				  , Next
				>
				type;
	};	

	template <class new_TaggedArg/*, class arg_list_T*/>							   //key was not found, so lets add it
	struct replace_key_from_index_T<0/*-1*/, new_TaggedArg/*, arg_list_T*/>
	{
		//typedef arg_list_T type;
		typedef arg_list<new_TaggedArg, self> type;
	};

/*
	template <class new_TaggedArg>							   //key was not found, so lets add it
	struct replace_key_from_index_T<0xFFFFFFFF, new_TaggedArg>		 //some strange MSVC compiler behaviour causes index value of 0xFFFFFFFF
	{
		//typedef arg_list_T type;
		typedef arg_list<new_TaggedArg, self> type;
	};
*/




	template <int index, class oldlist_T>								//partial function template specialization is not allowed therefore we need to do this dance
	struct do_remove_key_from_index_O
	{
		inline static typename remove_key_from_index_T<index>::type
		apply(oldlist_T const &oldlist)							 	
		{
			return /*arg_list<
					typename oldlist_T::TaggedArg_T
				  , replace_key_from_index_T_worker<index - 1, new_TaggedArg, typename oldlist_T::Next_T>::type
				>	 */
				remove_key_from_index_T<index>::type
				(
					oldlist.arg
				  , Next::do_remove_key_from_index_O
					<
						index - 1
					  , Next /*, typename oldlist_T::Next_T*/
					>
					::apply													//++
					(
						static_cast<Next>(oldlist)
					)	
				);														///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}
	};
	
	
	template <class oldlist_T>
	struct do_remove_key_from_index_O<1/*0*/, oldlist_T>
	{
		//BOOST_STATIC_ASSERT(false);
	
		inline static Next
		apply(oldlist_T const &oldlist)				//replace takes place here
		{
			//return static_cast<Next>(*this);
			return static_cast<Next>(oldlist);							///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}	
	};
	
	
	template <class oldlist_T>
	struct do_remove_key_from_index_O<0/*-1*/, oldlist_T>
	{
		inline static oldlist_T
		apply(oldlist_T const &oldlist)				//key was not found, so lets add it
		{
			//return *this;
			return oldlist;			///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}
	};








	template <int index, class new_TaggedArg, class oldlist_T>								//partial function template specialization is not allowed therefore we need to do this dance
	struct do_replace_key_from_index_O
	{
		inline static typename replace_key_from_index_T<index, new_TaggedArg>::type
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)							 	
		{
			return /*arg_list<
					typename oldlist_T::TaggedArg_T
				  , replace_key_from_index_T_worker<index - 1, new_TaggedArg, typename oldlist_T::Next_T>::type
				>	 */
				replace_key_from_index_T<index, new_TaggedArg>::type
				(
					oldlist.arg
				  , Next::do_replace_key_from_index_O
					<
						index - 1
					  , new_TaggedArg
					  , Next /*, typename oldlist_T::Next_T*/
					>
					::apply													//++
					(
						new_arg
					  , static_cast<Next>(oldlist)
					)	
				);														///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}
	};
	
	
	template <class new_TaggedArg, class oldlist_T>
	struct do_replace_key_from_index_O<1/*0*/, new_TaggedArg, oldlist_T>
	{
		//BOOST_STATIC_ASSERT(false);
	
		inline static arg_list<new_TaggedArg, Next>
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)				//replace takes place here
		{
			//return static_cast<Next>(*this);
			return arg_list<            
				new_TaggedArg 
			  , Next
			>(
				new_arg
			  , static_cast<Next>(oldlist)							///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
			);
		}	
	};
	
	
	template <class new_TaggedArg, class oldlist_T>
	struct do_replace_key_from_index_O<0/*-1*/, new_TaggedArg, oldlist_T>
	{
		inline static arg_list<new_TaggedArg, self>
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)				//key was not found, so lets add it
		{
			//return *this;
			return arg_list<new_TaggedArg, self>(new_arg, oldlist);			///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}
	};
/*	
	template <class new_TaggedArg, class oldlist_T>
	struct do_replace_key_from_index_O<0xFFFFFFFF, new_TaggedArg, oldlist_T>		 //some strange MSVC compiler behaviour causes index value of 0xFFFFFFFF
	{
		inline static arg_list<new_TaggedArg, self>
		apply(new_TaggedArg new_arg, oldlist_T const &oldlist)				//key was not found, so lets add it
		{
			//return *this;
			return arg_list<new_TaggedArg, self>(new_arg, oldlist);			///we must use oldlist param because static member functions do not have 'this' pointers. Static member function is neccessary because partial member function specialization is not available.
		}
	};
*/	
	
	
	

#endif	
	
	




	template <int index, class new_TaggedArg>
	inline typename replace_key_from_index_T<index, new_TaggedArg>::type	//inline by roland
	replace_key_from_index_O(new_TaggedArg new_arg) const							 	
	{	
		return do_replace_key_from_index_O<index, new_TaggedArg, self>::apply(new_arg, *this);	
	}

	template <int index>
	inline typename remove_key_from_index_T<index>::type	//inline by roland
	remove_key_from_index_O() const							 	
	{	
		return do_remove_key_from_index_O<index, self>::apply(*this);	
	}


	/*
	template <int index, class new_TaggedArg>								//partial function template specialization is not allowed therefore we need to do this dance above instead
	typename replace_key_from_index_T<index, new_TaggedArg>::type
	replace_key_from_index_O(new_TaggedArg new_arg) const							 	
	{
		return 
		replace_key_from_index_T<index>::type
		(
			arg
		  , Next::replace_key_from_index_O<index - 1, new_TaggedArg>(new_arg)	
		);
	}
	
	template <class new_TaggedArg>
	arg_list<new_TaggedArg, Next>
	replace_key_from_index_O<0>(new_TaggedArg new_arg) const				//replace takes place here
	{
		//return static_cast<Next>(*this);
		return arg_list<            
            new_TaggedArg 
          , Next
        >(
            new_arg
          , static_cast<Next>(*this)
        );
	}	
	
	template <class new_TaggedArg>
	arg_list<new_TaggedArg, self>
	replace_key_from_index_O<-1>(new_TaggedArg new_arg) const			//key was not found, so lets add it
	{
		//return *this;
		return arg_list<new_TaggedArg, self>(new_arg, *this);
	}
	*/
	
#endif	
	
	

	
		

    //roland end - if new param has same key as existing somewhere in current list, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous




    //roland start - allows merging argument lists. If new param has same key as current list head, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous


#if 0

	template <class KW, class T2>
    arg_list<tagged_argument<KW, T2>, typename remove_key_from_index_T<index_of_key<KW>::rval>::type> 
    operator,(tagged_argument<KW,T2> x) const				//roland - existing argument list is always const
    {        
        return arg_list<
			tagged_argument<KW,T2>
		  , typename remove_key_from_index_T<index_of_key<KW>::rval + 1>::type
		>(
			x
		  , replace_key_from_index_O<index_of_key<KW>::rval>()
		);
    }

#else						//roland - lets optimize so that new key and value are substituted into old one's memory slot. This way we prevent shifting the whole list around in memory





	template <class KW, class T2>   
	struct add_param
	{
		typedef typename replace_key_from_index_T<index_of_key<KW>::rval + 1, tagged_argument<KW,T2> >::type		type;		   //+1 is neccessary because MSVC release build for some reason does not match template of -1, it sees instead 0xFFFFFFFF
	};






	template <class KW, class T2>
	inline																 //inline by roland
    typename add_param<KW, T2>::type			    
    operator,(tagged_argument<KW,T2> x) const				//roland - existing argument list is always const
    {   
        return replace_key_from_index_O<index_of_key<KW>::rval + 1, tagged_argument<KW,T2> >(x);		  //+1 is neccessary because MSVC release build for some reason does not match template of -1, it sees instead 0xFFFFFFFF
    }
   
   
    
   
   
	//this method is useful for case when some macro sometimes adds a parameter and sometimes not (using then empty_arg_list instead), but we still want to use comma before that macro
	inline const self&																 //inline by roland
    operator,(empty_arg_list x) const				//roland - existing argument list is always const
    {   
        return *this;		  
    }
 
 
 
   
 
    
    
    //TODO: merging arg lists. ex: (P2 = 3, P4 = b(), p)		where p is called as (P4 = a())
    
    

#endif


    //roland end - allows merging argument lists. If new param has same key as current list head, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous

#endif    








    //roland start - isset, unset and set methods
    
    template <class key>
	bool isset2(key &) const
	{
		return (index_of_keyword<key>::rval == -1) ? false : true;			//NB! index_of_keyword here, not index_of_key
	}
	
	template <class key>
	struct isset
	{
		static const bool rval = (index_of_key<key>::rval == -1) ? false : true;
	};
        
        
        
        
        
        

	template <class key>													 ///prevents Error C2785 'declaration1' and 'declaration2' have different return types for unset method
	struct get_unset_rT
	{
		//typedef arg_list<tagged_argument<Keyword, Arg>, empty_arg_list>		type;	
		static const int index												= index_of_keyword<key>::rval;	
		typedef typename remove_key_from_index_T<index + 1>::type			type;								//+1 is neccessary because MSVC release build for some reason does not match template of -1, it sees instead 0xFFFFFFFF
	};
		
	
    
    template <class key>
	typename get_unset_rT<key>::type
	unset(key &) const
	{
		static const int index												= index_of_keyword<key>::rval;	
		
		return remove_key_from_index_O<index + 1>();				//+1 is neccessary because MSVC release build for some reason does not match template of -1, it sees instead 0xFFFFFFFF
	};
		
        
        
    //roland end - isset, unset and set methods
    

    
    
    

    
    
};

#pragma warning (pop)						//roland - warning C4512: 'boost::parameter::aux::arg_list<TaggedArg,Next>' : assignment operator could not be generated	    



#if BOOST_WORKAROUND(BOOST_MSVC, <= 1300)  // ETI workaround
template <> struct arg_list<int,int> {};
#endif 

}}} // namespace boost::parameter::aux

#endif // ARG_LIST_050329_HPP

